#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import abc
import csv
from typing import IO, Callable, Iterator, Self, Sequence, Type

from openepd.bundle.model import AssetInfo, AssetType, BundleManifest, RelType
from openepd.model.base import TOpenEpdObject

AssetFilter = Callable[[AssetInfo], bool]
"""A filter function for assets. Returns True if the asset should be included."""
AssetRef = str | AssetInfo


class toc_dialect(csv.Dialect):
    """Describe the usual properties of Excel-generated CSV files."""

    delimiter = ","
    quotechar = '"'
    doublequote = True
    skipinitialspace = True
    lineterminator = "\r\n"
    quoting = csv.QUOTE_NONNUMERIC


csv.register_dialect("toc", toc_dialect)


class BundleMixin:
    """Mixin for bundle readers and writers."""

    _TOC_FIELDS: tuple[str, ...] = (
        "ref",
        "name",
        "type",
        "lang",
        "rel_type",
        "rel_asset",
        "comment",
        "content_type",
        "size",
        "custom_type",
        "custom_data",
    )

    @classmethod
    def _asset_ref_to_str(cls, asset_ref: AssetRef) -> str:
        if isinstance(asset_ref, AssetInfo):
            return asset_ref.ref
        else:
            return asset_ref


class BaseBundleReader(BundleMixin, metaclass=abc.ABCMeta):
    """Base class for bundle readers."""

    def __enter__(self) -> Self:
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    @abc.abstractmethod
    def get_manifest(self) -> BundleManifest:
        """Get the manifest of the bundle. Manifest object is immutable."""
        pass

    @abc.abstractmethod
    def close(self):
        """Close the reader."""
        pass

    @abc.abstractmethod
    def get_relative_assets_iter(
        self, asset: AssetRef, rel_type: str | Sequence[str] | None = None
    ) -> Iterator[AssetInfo]:
        """
        Get all assets that are related to the given asset.

        :param asset: The asset to get related assets for.
        :param rel_type: The type of the relation. If None, all relations are returned.
        :return: An iterator over the related assets.
        """
        pass

    @abc.abstractmethod
    def assets_iter(self) -> Iterator[AssetInfo]:
        """Get an iterator over all assets in the bundle."""
        pass

    @abc.abstractmethod
    def root_assets_iter(
        self,
        filter_or_type: AssetFilter | str | AssetType | None = None,
        name: str | None = None,
        parent_ref: AssetRef | None = None,
        ref_type: str | None = None,
        is_translated: bool | None = None,
    ) -> Iterator[AssetInfo]:
        """
        Get an iterator over all root assets in the bundle. Optionally, filter the assets.

        :param filter_or_type: A filter function or an asset type to filter by.
        :param name: The name of the asset to filter by.
        :param parent_ref: The parent asset to filter by.
        :param ref_type: The type of the asset to filter by.
        :param is_translated: Whether the asset is translated.
        :return: An iterator over the root assets.
        """
        pass

    @abc.abstractmethod
    def get_asset_by_ref(self, asset_ref: AssetRef) -> AssetInfo | None:
        """Get an asset by its reference or None if not found."""
        pass

    @abc.abstractmethod
    def read_blob_asset(self, asset_ref: AssetRef) -> IO[bytes]:
        """Read a blob asset by given reference."""
        pass

    @abc.abstractmethod
    def read_object_asset(self, obj_class: Type[TOpenEpdObject], asset_ref: AssetRef) -> TOpenEpdObject:
        """Read an object asset by given reference."""
        pass

    def get_relative_assets(self, asset: AssetInfo, rel_type: str | Sequence[str] | None = None) -> list[AssetInfo]:
        """Get all assets that are related to the given asset."""
        return list(self.get_relative_assets_iter(asset, rel_type))

    def get_first_relative_asset(
        self, asset: AssetInfo, rel_type: str | Sequence[str] | None = None
    ) -> AssetInfo | None:
        """
        Get the first asset that is related to the given asset or None if not found.

        :param asset: The asset to get related assets for.
        :param rel_type: The type of the relation. If None, all relations are returned.
        """
        for x in self.get_relative_assets_iter(asset, rel_type):
            return x
        return None

    def get_translations_for_asset(self, asset: AssetRef) -> list[AssetInfo]:
        """Get all translations for the given asset."""
        return list(self.get_relative_assets_iter(asset, rel_type=RelType.Translation))

    def get_first_root_asset(self, asset_type: AssetType) -> AssetInfo | None:
        """Get the first root asset of the given type or None if not found."""
        for x in self.root_assets_iter(asset_type):
            return x
        return None


class BaseBundleWriter(BundleMixin, metaclass=abc.ABCMeta):
    """Base class for bundle writers."""

    def __enter__(self) -> Self:
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    @abc.abstractmethod
    def write_blob_asset(
        self,
        data: IO[bytes],
        content_type: str | None,
        rel_asset: AssetRef | None = None,
        rel_type: str | None = None,
        file_name: str | None = None,
        name: str | None = None,
        lang: str | None = None,
        comment: str | None = None,
        custom_type: str | None = None,
        custom_data: str | None = None,
    ) -> AssetInfo:
        """Write a blob asset."""
        pass

    @abc.abstractmethod
    def write_object_asset(
        self,
        obj: TOpenEpdObject,
        rel_asset: AssetRef | None = None,
        rel_type: str | None = None,
        file_name: str | None = None,
        name: str | None = None,
        lang: str | None = None,
        comment: str | None = None,
        custom_type: str | None = None,
        custom_data: str | None = None,
    ) -> AssetInfo:
        """Write an object asset."""
        pass

    @abc.abstractmethod
    def commit(self):
        """Write all relevant metadata into the bundle."""
        pass

    @abc.abstractmethod
    def close(self):
        """Close the writer."""
        pass



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Any


class ApiError(Exception):
    """Base class for all API errors."""

    def __init__(
        self, http_status: int, error_summary: str, response: Any = None, error_code: str | None = None
    ) -> None:
        super().__init__(error_summary)
        self.error_summary = error_summary
        self.http_status = http_status
        self.error_code = error_code
        self.response = response


class ObjectNotFound(ApiError):
    """Object not found error."""

    pass


class ServerError(ApiError):
    """Server error."""

    pass


class ValidationError(ApiError):
    """Validation error."""

    def __init__(
        self,
        http_status: int,
        error_summary: str,
        validation_errors: dict[str, list[str]] | None,
        response: Any,
        error_code: str | None,
    ) -> None:
        super().__init__(http_status, error_summary, response, error_code)
        self.validation_errors: dict[str, list[str]] = validation_errors or {}

    @staticmethod
    def __flatten(nested_errors: dict[str, Any], parent_key: str = "") -> dict[str, list[str]]:
        result: dict[str, list[str]] = {}
        for key, value in nested_errors.items():
            full_key = f"{parent_key}.{key}" if parent_key else key
            if isinstance(value, dict):
                nested_result = ValidationError.__flatten(value, full_key)
                for nested_key, nested_errors in nested_result.items():  # type: ignore[assignment]
                    result.setdefault(nested_key, []).extend(nested_errors)
            else:
                result.setdefault(full_key, []).extend(value)
        return result

    def __str__(self) -> str:
        result: list[str] = ["Validation errors:"]
        for code, errors in self.__flatten(self.validation_errors).items():  # type: ignore[arg-type]
            result.append(f"{code}:")
            for e in errors:
                result.append(f"  {e}")
        return "\n".join(result)


class AuthError(ApiError):
    """Common class for any authentication related errors."""

    pass


class NotAuthorizedError(AuthError):
    """Not authorized error."""

    pass


class AccessDeniedError(AuthError):
    """Access denied error."""

    pass



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import (
    FlatGlassPanesThickness,
    FrameMaterial,
    HardwareFunction,
    NAFSPerformanceGrade,
    Spacer,
    ThermalSeparation,
)
from openepd.model.validation.numbers import RatioFloat
from openepd.model.validation.quantity import (
    LengthMmStr,
    PressureMPaStr,
    SpeedStr,
    UFactorStr,
    validate_quantity_unit_factory,
)


class GlazingIntendedApplication(BaseOpenEpdSchema):
    """Glass intended application mixin."""

    curtain_wall: bool | None = pyd.Field(default=None, description="Intended for curtain walls. Relevant for IGUs.")
    r_windows: bool | None = pyd.Field(
        default=None,
        description="Intended for residential (NAFS 'R') and similar windows, doors, or skylights. Relevant for IGUs.",
    )
    lc_windows: bool | None = pyd.Field(
        default=None, description="Intended for light commercial (NAFS 'LC') and similar windows. Relevant for IGUs."
    )
    cw_windows: bool | None = pyd.Field(
        default=None, description="Intended for commercial (NAFS 'CW') and similar windows. Relevant for IGUs."
    )
    aw_windows: bool | None = pyd.Field(
        default=None, description="Intended for architectural (NAFS 'AW') and similar windows. Relevant for IGUs."
    )
    storefronts: bool | None = pyd.Field(
        default=None, description="Intended for Storefronts and similar applications. Relevant for IGUs."
    )
    glazed_doors: bool | None = pyd.Field(
        default=None, description="Intended for Glazed Doors and similar applications. Relevant for IGUs."
    )
    unit_skylights: bool | None = pyd.Field(
        default=None, description="Intended for Unit Skylights and similar applications. Relevant for IGUs."
    )
    sloped_skylights: bool | None = pyd.Field(
        default=None,
        description="Intended for sloped glazing, and architectural skylights, and similar. Relevant for IGUs.",
    )
    other: bool | None = pyd.Field(
        default=None, description="Intended for other application not listed. Relevant for IGUs."
    )


class GlazingOptionsMixin(BaseOpenEpdSchema):
    """Common glazing options."""

    low_emissivity: bool | None = pyd.Field(default=None, description="Low Emissivity coatings", example=True)
    electrochromic: bool | None = pyd.Field(
        default=None,
        description="Glazing with an electrically controllable solar heat gain and/or other properties.",
        example=True,
    )
    acid_etched: bool | None = pyd.Field(
        default=None, description="Flat glass that has undergone a chemical etching process.", example=True
    )
    tempered: bool | None = pyd.Field(
        default=None,
        description="Consists of a single pane that has been heat-treated to give the glass increased impact "
        "resistance. Standard typically used in North America.",
        example=True,
    )
    toughened: bool | None = pyd.Field(
        default=None,
        description="Consists of a single pane that has been specially heat-treated to give the glass increased impact "
        "resistance. Standard typically used in Europe.",
        example=True,
    )
    laminated: bool | None = pyd.Field(
        default=None,
        description="Consists of at least two glass panes lying one on top of the other, with one or several layers of "
        "a tear-resistant, viscoelastic film positioned between the panes, which consist of polyvinyl "
        "butyral (PVB)",
        example=True,
    )
    fire_resistant: bool | None = pyd.Field(default=None, example=True)
    fire_protection: bool | None = pyd.Field(
        default=None,
        description="Specifically tested for its ability to block flames and smoke, but not radiant heat. Ranges from"
        " specialty tempered products rated for ~20 minutes to glass ceramics rated up to 3 hours.",
        example=True,
    )
    pyrolytic_coated: bool | None = pyd.Field(
        default=None,
        description="At least one coating is applied in a pyrolytic process, typically during float glass production.",
        example=True,
    )
    sputter_coat: bool | None = pyd.Field(
        default=None,
        description="At least one coating is applied using sputter (vacuum deposition) coating.",
        example=True,
    )
    solar_heat_gain: RatioFloat | None = pyd.Field(
        default=None,
        description="Solar heat gain, measured at a certain level of Differential Pressure. Range is 0 to 1.",
        example=0.5,
        ge=0,
        le=1,
    )


class PanelDoorsV1(BaseOpenEpdHierarchicalSpec):
    """Panel doors performance specification."""

    _EXT_VERSION = "1.0"


class PressureResistantDoorsV1(BaseOpenEpdHierarchicalSpec):
    """Pressure-Resistant Doors."""

    _EXT_VERSION = "1.0"


class SpecialFunctionDoorsV1(BaseOpenEpdHierarchicalSpec):
    """
    Special function doors.

    Includes doors for e.g., cold storage, hangars, lightproof applications,
    security, sound control, vaults, etc.
    """

    _EXT_VERSION = "1.0"

    # Nested specs:
    PanelDoors: PanelDoorsV1 | None = None
    PressureResistantDoors: PressureResistantDoorsV1 | None = None


class SlidingGlassDoorsV1(BaseOpenEpdHierarchicalSpec):
    """Sliding glass doors performance specification."""

    _EXT_VERSION = "1.0"


class FenestrationAccessoriesV1(BaseOpenEpdHierarchicalSpec):
    """
    Fenestration accessories.

    Gaskets, seals, fasteners, and other low-mass items which may be useful in calculating the impact of a
    fenestration system.
    """

    _EXT_VERSION = "1.0"


class FenestrationFramingV1(BaseOpenEpdHierarchicalSpec):
    """
    Fenestration Framing.

    Lineal elements ("sticks") for use in fenestration, including frames, sashes, and mullions.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    thermal_separation: ThermalSeparation | None = pyd.Field(default=None, example="Aluminium")
    material: FrameMaterial | None = pyd.Field(default=None, example="Vinyl")


class FenestrationHardwareV1(BaseOpenEpdHierarchicalSpec):
    """Locks, operation hardware, and other substantial items declared on a per-piece basis."""

    _EXT_VERSION = "1.0"

    # Own fields:
    function: HardwareFunction | None = pyd.Field(default=None, description="", example="Lock")


class FlatGlassPanesV1(BaseOpenEpdHierarchicalSpec):
    """Monolithic, uncoated flat glass panes that are not substantially processed."""

    _EXT_VERSION = "1.0"

    # Own fields:
    thickness: FlatGlassPanesThickness | None = pyd.Field(default=None, example="12 mm")

    _flat_glass_panes_thickness_is_quantity_validator = pyd.validator("thickness", allow_reuse=True)(
        validate_quantity_unit_factory("m")
    )


class ProcessedNonInsulatingGlassPanesV1(BaseOpenEpdHierarchicalSpec, GlazingOptionsMixin):
    """
    Solid glass panes without internal gaps which have been heat-treated or otherwise substantially processed.

    Includes:
    1. Coatings including low-e and other coatings (see PCR)
    2. laminating (fire-rated, glass clad polycarbonate, interlayers
    3. Heat treated (heat strengthened, tempered, fire-rated)
    4. Mechanically or chemically processed or fabricated
    (edging, bending, etching, drilling, notching, cutting, polishing, etc)
    5. combined products of processing in 1-5.
    """

    _EXT_VERSION = "1.0"


class GlazedDoorsV1(BaseOpenEpdHierarchicalSpec):
    """
    Factory assembled door which is at least 50% glass by area.

    Includes sliding patio doors and hinged doors.
    """

    _EXT_VERSION = "1.0"


class UnitSkylightsV1(BaseOpenEpdHierarchicalSpec):
    """
    Unit skylights performance specification.

    A factory assembled fenestration unit for installation on the roof of a structure to provide interior
    building spaces with natural daylight, warmth, and ventilation; generally not operable
    by hand (cf. roof window). Includes frame(s) and possibly operating hardware.
    """

    _EXT_VERSION = "1.0"


class WindowsV1(BaseOpenEpdHierarchicalSpec):
    """Windows including glazing and frame material."""

    _EXT_VERSION = "1.0"


class IntegratedDoorsOpeningAssembliesV1(BaseOpenEpdHierarchicalSpec):
    """Pre-installed unit that includes door, frame, and hardware."""

    _EXT_VERSION = "1.0"


class MetalDoorAndFramesV1(BaseOpenEpdHierarchicalSpec):
    """Metal doors and frames."""

    _EXT_VERSION = "1.0"


class SpecialtyDoorsAndFramesV1(BaseOpenEpdHierarchicalSpec):
    """
    Specialty doors and frames.

    Includes e.g., access doors and panels, sliding glass doors, coiling doors, special function doors,
    folding doors, etc.
    """

    _EXT_VERSION = "1.0"

    # Nested specs:
    SpecialFunctionDoors: SpecialFunctionDoorsV1 | None = None
    SlidingGlassDoors: SlidingGlassDoorsV1 | None = None


class WoodDoorsV1(BaseOpenEpdHierarchicalSpec):
    """Wood doors performance specification."""

    _EXT_VERSION = "1.0"


class FenestrationPartsV1(BaseOpenEpdHierarchicalSpec):
    """
    Fenestration Parts.

    Parts and assemblies for integration into building fenestration such as windows, curtain walls,
    and storefronts.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    intended_application: GlazingIntendedApplication | None = pyd.Field(
        default=None, description="Intended application."
    )

    # Nested specs:
    FenestrationAccessories: FenestrationAccessoriesV1 | None = None
    FenestrationFraming: FenestrationFramingV1 | None = None
    FenestrationHardware: FenestrationHardwareV1 | None = None


class GlassPanesV1(BaseOpenEpdHierarchicalSpec):
    """Flat glass panes."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    FlatGlassPanes: FlatGlassPanesV1 | None = None
    ProcessedNonInsulatingGlassPanes: ProcessedNonInsulatingGlassPanesV1 | None = None


class NAFSPerformanceClass(BaseOpenEpdSchema):
    """NAFS Performance class."""

    r: bool | None = pyd.Field(
        default=None, description="Residential; commonly used in one- and two-family dwellings.", example=True
    )
    lc: bool | None = pyd.Field(
        default=None,
        description="Light Commercial: commonly used in low-rise and mid-rise multi-family dwellings and other "
        "buildings where larger sizes and higher loading requirements are expected.",
        example=True,
    )
    cw: bool | None = pyd.Field(
        default=None,
        description="Commercial Window: commonly used in low-rise and mid-rise buildings where larger sizes, higher "
        "loading requirements, limits on deflection, and heavy use are expected.",
        example=True,
    )
    aw: bool | None = pyd.Field(
        default=None,
        description="Architectural Window: commonly used in high-rise and mid-rise buildings to meet increased "
        "loading requirements and limits on deflection, and in buildings where frequent and extreme use "
        "of the fenestration products is expected.",
        example=True,
    )


class NAFSFenestrationV1(BaseOpenEpdHierarchicalSpec, GlazingOptionsMixin):
    """Factory assembled fenestration units compliant to the North American Fenestration Standard."""

    _EXT_VERSION = "1.0"

    # Own fields:
    hurricane_resistant: bool | None = pyd.Field(
        default=None, description="The product has been designed to resist windborne debris.", example=True
    )

    assembly_u_factor: UFactorStr | None = pyd.Field(
        default=None,
        description="Weighted average conductance of heat across assembly (including frame).",
        example="1 USI",
    )
    air_infiltration: SpeedStr | None = pyd.Field(
        default=None,
        description="Air infiltration, measured at a certain level of Differential Pressure.",
        example="1 m3 / m2 / s",
    )

    thermal_separation: ThermalSeparation | None = pyd.Field(default=None, example="Aluminium")
    dp_rating: PressureMPaStr | None = pyd.Field(default=None, description="", example="1 MPa")
    glass_panes: int | None = pyd.Field(
        default=None,
        description="Number of panes, each separated by a cavity. A 3 pane unit has 2 cavities. example: 3",
        example=3,
    )

    performance_class: NAFSPerformanceClass | None = pyd.Field(
        default=None, description="Performance class according to NAFS."
    )

    performance_grade: NAFSPerformanceGrade | None = pyd.Field(
        default=None,
        description="NAFS Performance Grade. The NAFS Performance Grade is a number that represents the performance "
        "of the glazing product. The higher the number, the better the performance. The NAFS Performance "
        "Grade is calculated using the NAFS Performance Class, the NAFS Performance Index, and the NAFS "
        "Performance Factor. While it is expressed as pressure, there are specific values which are "
        "allowed. The values are listed in the enum.",
        example="95 psf",
    )

    _assembly_u_factor_is_quantity_validator = pyd.validator("assembly_u_factor", allow_reuse=True)(
        validate_quantity_unit_factory("USI")
    )
    _nafs_performance_grade_is_quantity_validator = pyd.validator("performance_grade", allow_reuse=True)(
        validate_quantity_unit_factory("psf")
    )

    # Nested specs:
    GlazedDoors: GlazedDoorsV1 | None = None
    UnitSkylights: UnitSkylightsV1 | None = None
    Windows: WindowsV1 | None = None


class InsulatingGlazingUnitsV1(BaseOpenEpdHierarchicalSpec, GlazingOptionsMixin):
    """Insulating glazing units performance specification."""

    _EXT_VERSION = "1.0"

    # Own fields:
    intended_application: GlazingIntendedApplication | None = pyd.Field(
        default=None, description="Intended application for IGUs."
    )

    hurricane_resistant: bool | None = pyd.Field(default=None, example=True)

    dp_rating: PressureMPaStr | None = pyd.Field(
        default=None, description="Maximum Differential Pressure, a measure of wind tolerance.", example="1 MPa"
    )
    air_infiltration: SpeedStr | None = pyd.Field(
        default=None,
        description="Air infiltration, measured at a certain level of Differential Pressure.",
        example="1 m3 / m2 / s",
    )
    glass_panes: int | None = pyd.Field(
        default=None,
        description="Number of panes, each separated by a cavity. A 3 pane unit has 2 cavities. example: 3",
        example=3,
    )
    cog_u_factor: UFactorStr | None = pyd.Field(
        default=None, description="Conductance of heat at center of glass.", example="1 USI"
    )
    spacer: Spacer | None = pyd.Field(
        default=None, description="Spacer material for Integrated Glass Unit.", example="Aluminium"
    )

    _dp_rating_is_quantity_validator = pyd.validator("dp_rating", allow_reuse=True)(
        validate_quantity_unit_factory("MPa")
    )
    _cog_u_factor_is_quantity_validator = pyd.validator("cog_u_factor", allow_reuse=True)(
        validate_quantity_unit_factory("USI")
    )


class CurtainWallsV1(BaseOpenEpdHierarchicalSpec):
    """
    Curtain Walls.

    Exterior skin of building where walls are non-structural and are outboard of the floor slabs,
    often as system of aluminum framing with vision glass and opaque panels of glass, metal, or other
    materials.

    Can be 'unitized' (prefabricated off-site) or 'stick' (fabricated on site).
    """

    _EXT_VERSION = "1.0"


class DoorsAndFramesV1(BaseOpenEpdHierarchicalSpec):
    """Doors (the operable part) and frames (what holds the door proper)."""

    _EXT_VERSION = "1.0"

    # Own fields:
    height: LengthMmStr | None = pyd.Field(default=None, example="1200 mm")
    width: LengthMmStr | None = pyd.Field(default=None, example="600 mm")

    # Nested specs:
    IntegratedDoorsOpeningAssemblies: IntegratedDoorsOpeningAssembliesV1 | None = None
    MetalDoorAndFrames: MetalDoorAndFramesV1 | None = None
    SpecialtyDoorsAndFrames: SpecialtyDoorsAndFramesV1 | None = None
    WoodDoors: WoodDoorsV1 | None = None


class EntrancesV1(BaseOpenEpdHierarchicalSpec):
    """Building entrances (distinct from the door proper)."""

    _EXT_VERSION = "1.0"


class GlazingV1(BaseOpenEpdHierarchicalSpec):
    """
    Glazing performance specification.

    Broad category of glass-based products, accessories, and assemblies ranging from
    glass panes and framing to curtain walls.
    """

    _EXT_VERSION = "1.0"

    # Nested specs:
    FenestrationParts: FenestrationPartsV1 | None = None
    GlassPanes: GlassPanesV1 | None = None
    NAFSFenestration: NAFSFenestrationV1 | None = None
    InsulatingGlazingUnits: InsulatingGlazingUnitsV1 | None = None


class StorefrontsV1(BaseOpenEpdHierarchicalSpec):
    """
    Storefronts.

    Fabricated building facades commonly used in retail applications, typically one or two stories
    tall and using metal framing and glass.
    """

    _EXT_VERSION = "1.0"


class TranslucentWallAndRoofAssembliesV1(BaseOpenEpdHierarchicalSpec):
    """
    Translucent wall and roof assemblies.

    Includes structured polycarbonate panel and fiberglass sandwich panel assemblies.
    """

    _EXT_VERSION = "1.0"


class WindowWallAssembliesV1(BaseOpenEpdHierarchicalSpec):
    """
    Window Wall Assemblies.

    Exterior skin of building where walls are non-structural and sit between floor slabs, often as
    system of aluminum framing with vision glass and opaque panels of glass, metal,
    or other materials.
    """

    _EXT_VERSION = "1.0"


class OpeningsV1(BaseOpenEpdHierarchicalSpec):
    """
    Openings performance specification.

    General category that includes windows, storefronts, window walls, curtain walls,
    doors, entrances, etc.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    thickness: LengthMmStr | None = pyd.Field(default=None, example="80 mm")

    # Nested specs:
    CurtainWalls: CurtainWallsV1 | None = None
    DoorsAndFrames: DoorsAndFramesV1 | None = None
    Entrances: EntrancesV1 | None = None
    Glazing: GlazingV1 | None = None
    Storefronts: StorefrontsV1 | None = None
    TranslucentWallAndRoofAssemblies: TranslucentWallAndRoofAssembliesV1 | None = None
    WindowWallAssemblies: WindowWallAssembliesV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = ("AsphaltRangeV1",)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from openepd.compat.pydantic import pyd
from openepd.model.common import RangeRatioFloat
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import AsphaltGradation, AsphaltMixType
from openepd.model.validation.quantity import AmountRangeLengthMm, TemperatureCStr


class AsphaltRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    General category for asphalt mixtures.

    Range version.
    """

    _EXT_VERSION = "1.0"

    aggregate_size_max: AmountRangeLengthMm | None = pyd.Field(default=None, description="Max aggregate size")
    rap: RangeRatioFloat | None = pyd.Field(
        default=None, description="Percent of mixture that has been replaced by recycled asphalt pavement (RAP)."
    )
    ras: RangeRatioFloat | None = pyd.Field(
        default=None, description="Percent of mixture that has been replaced by recycled asphalt shingles (RAS)."
    )
    ground_tire_rubber: RangeRatioFloat | None = pyd.Field(
        default=None, description="Percent of mixture that has been replaced by ground tire rubber (GTR)."
    )
    max_temperature: TemperatureCStr | None = pyd.Field(
        default=None,
        description="The upper threshold temperature to which an asphalt binder can be heated preventing the asphalt mixture from rutting",
    )
    min_temperature: TemperatureCStr | None = pyd.Field(
        default=None,
        description="The lower threshold temperature for an asphalt binder to prevent thermal cracking of the asphalt mixture.",
    )
    mix_type: list[AsphaltMixType] | None = pyd.Field(default=None, description="Asphalt mix type")
    gradation: list[AsphaltGradation] | None = pyd.Field(default=None, description="Asphalt gradation")
    sbr: bool | None = pyd.Field(default=None, description="Styrene-butadiene rubber (SBR)")
    sbs: bool | None = pyd.Field(default=None, description="Styrene-butadiene-styrene (SBS)")
    ppa: bool | None = pyd.Field(default=None, description="Polyphosphoric acid (PPA)")
    gtr: bool | None = pyd.Field(default=None, description="Ground tire rubber (GTR)")
    pmb: bool | None = pyd.Field(default=None, description="Polymer modified bitumen (PMB)")



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.concrete import Cementitious
from openepd.model.specs.enums import (
    AdmixtureEffects,
    CarpetYarnType,
    CementAstmType,
    CementC1157,
    CementCsaA3001,
    CementEn197_1,
    CementScm,
    MasonryCementAstmC91Type,
    TextilesFabricType,
)


class CementV1(BaseOpenEpdHierarchicalSpec):
    """
    Cements.

    Includes Portland and blended cements, that can serve as the primary binder in a concrete mix.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    cementitious: Cementitious | None = pyd.Field(default=None, description="")
    white_cement: bool | None = pyd.Field(default=None, description="", example=True)
    astm_type: CementAstmType | None = pyd.Field(default=None, description="", example="C150 Type I")
    c1157: list[CementC1157] | None = pyd.Field(default=None, description="", example=["GU"])
    csa_a3001: list[CementCsaA3001] | None = pyd.Field(default=None, description="", example=["A3001 GU"])
    en197_1: CementEn197_1 | None = pyd.Field(default=None, description="", example="CEM I")
    oil_well_cement: bool | None = pyd.Field(default=None, description="", example=True)


class MasonryCementV1(BaseOpenEpdHierarchicalSpec):
    """
    A cementitious product typically composed of Portland cement and hydrated lime.

    Masonry cement is combined with sand to make mortar.
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    astm_c91_type: MasonryCementAstmC91Type | None = pyd.Field(default=None, description="", example="Type N")


class SupplementaryCementitiousMaterialsV1(BaseOpenEpdHierarchicalSpec):
    """Cementitious materials that are not effective binders when used on their own."""

    _EXT_VERSION = "1.0"

    # Own fields:
    cement_scm: list[CementScm] | None = pyd.Field(default=None, description="", example=["ggbs"])


class AccessFlooringPedestalsV1(BaseOpenEpdHierarchicalSpec):
    """
    Part of an access floor system.

    Pedestals are laid out on top of a floor slab and support access floor panels, creating a void space between the
    floor slab and finish floor.
    """

    _EXT_VERSION = "1.0"


class CarpetBackingV1(BaseOpenEpdHierarchicalSpec):
    """Fabric backing holding a carpet together."""

    _EXT_VERSION = "1.0"


class CarpetFiberV1(BaseOpenEpdHierarchicalSpec):
    """Fiber yarn used in the manufacture of carpet, typically nylon or wool."""

    _EXT_VERSION = "1.0"

    # Own fields:
    yarn_material: CarpetYarnType | None = pyd.Field(default=None, description="", example="Nylon 6,6")
    yarn_recycled_content: float | None = pyd.Field(default=None, description="", example=2.3)


class CementitiousMaterialsV1(BaseOpenEpdHierarchicalSpec):
    """Cementitious materials performance specification."""

    _EXT_VERSION = "1.0"

    # Nested specs:
    Cement: CementV1 | None = None
    MasonryCement: MasonryCementV1 | None = None
    SupplementaryCementitiousMaterials: SupplementaryCementitiousMaterialsV1 | None = None


class ConcreteAdmixturesV1(BaseOpenEpdHierarchicalSpec):
    """
    Concrete admixtures.

    Chemical additives that are added to fresh concrete immediately before or during mixing. Admixtures have distinct
    functions and are categorized as: air-entraining, water-reducing, retarding, accelerating, and plasticizers
    (i.e., superplasticizers).
    """

    _EXT_VERSION = "1.0"

    # Own fields:
    effects: list[AdmixtureEffects] | None = pyd.Field(default=None, description="", example=["Air Entrainer"])


class TextilesV1(BaseOpenEpdHierarchicalSpec):
    """Textiles for use in manufacturing end products."""

    _EXT_VERSION = "1.0"

    # Own fields:
    fabric_type: list[TextilesFabricType] | None = pyd.Field(default=None, description="", example=["Leather"])


class ManufacturingInputsV1(BaseOpenEpdHierarchicalSpec):
    """
    Manufacturing inputs.

    Broad category for collecting materials primarily used as manufacturing inputs, rather than directly used in
    a construction.
    """

    _EXT_VERSION = "1.0"

    # Nested specs:
    AccessFlooringPedestals: AccessFlooringPedestalsV1 | None = None
    CarpetBacking: CarpetBackingV1 | None = None
    CarpetFiber: CarpetFiberV1 | None = None
    CementitiousMaterials: CementitiousMaterialsV1 | None = None
    ConcreteAdmixtures: ConcreteAdmixturesV1 | None = None
    Textiles: TextilesV1 | None = None



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = ["ConduitMixin"]

from openepd.compat.pydantic import pyd
from openepd.model.specs.base import BaseOpenEpdSpec
from openepd.model.specs.enums import ConduitMaterial
from openepd.model.validation.quantity import LengthMmStr


class ConduitMixin(BaseOpenEpdSpec):
    """
    Properties of a conduit.

    Those properties are the same for communication and electrical conduits.
    """

    nominal_diameter: LengthMmStr | None = pyd.Field(
        default=None,
        description="Nominal Diameter is also known as the mean or average outside diameter.",
        example="100 mm",
    )
    outer_diameter: LengthMmStr | None = pyd.Field(
        default=None,
        description="The measurement of the distance of a straight line between points on the outer walls of the pipe.",
        example="100 mm",
    )
    inner_diameter: LengthMmStr | None = pyd.Field(
        default=None,
        description="The measurement of the distance of a straight line between points on the inner walls of the pipe.",
        example="100 mm",
    )
    wall_thickness: LengthMmStr | None = pyd.Field(
        default=None, description="Conduit wall thickness.", example="100 mm"
    )

    material: ConduitMaterial | None = pyd.Field(default=None, description="Material of the conduit.", example="PVC")



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "HttpStreamReader",
    "SyncHttpClient",
    "DoRequest",
    "RetryHandler",
    "BaseApiMethodGroup",
    "USER_AGENT_DEFAULT",
)

import datetime
from functools import partial, wraps
from io import IOBase
import logging
import random
import shutil
import time
from typing import IO, Any, BinaryIO, Callable, Final, NamedTuple

import requests
from requests import PreparedRequest, Response, Session, Timeout
from requests import codes as requests_codes
from requests.auth import AuthBase
from requests.structures import CaseInsensitiveDict

from openepd.__version__ import VERSION
from openepd.api import errors
from openepd.api.common import Throttler, no_trailing_slash

logger = logging.getLogger(__name__)

USER_AGENT_DEFAULT: Final[str] = f"OpenEPD API Client/{VERSION}"

DoRequest = Callable[[], Response]
RetryHandler = Callable[[DoRequest], Response | None]
ErrorHandler = Callable[[Response, bool], Response | None]
"""
ErrorHandler is a callable that takes a response and a boolean indicating whether 
exception should be raised in case of error.
"""


class HttpStreamReader(IOBase):
    """A wrapper around a requests Response object that allows it to be used as a stream."""

    def __init__(self, http_response: Response, chunk_size: int = 1024) -> None:
        """
        Initialize HttpStreamReader with given HttpResponse in streaming mode.

        :param http_response: HTTP response object
        """
        super().__init__()
        self._http_response = http_response
        self.chunk_size = chunk_size

    def readable(self) -> bool:
        """Return True if the stream can be read from."""
        return True

    def seekable(self) -> bool:
        """Return True if the stream supports random access."""
        return False

    def writable(self) -> bool:
        """Return True if the stream supports writing."""
        return False

    def read(self, size: int = -1) -> bytes:
        """Read and return up to size bytes, where size is an int."""
        return self._http_response.raw.read(size)

    def readinto(self, target_stream: IO[bytes]) -> None:
        """Read bytes into a pre-allocated, writable bytes-like object target_stream."""
        shutil.copyfileobj(self._http_response.raw, target_stream)

    def raw(self) -> BinaryIO:
        """Return the underlying HTTP Response content."""
        return self._http_response.raw

    def detach(self) -> BinaryIO:
        """Separate the underlying HTTP Response from the HttpStreamReader and return it."""
        tmp = self._http_response
        self._http_response = None  # type: ignore
        return tmp.raw

    def close(self) -> None:
        """
        Releases the connection back to the pool.

        Once this method has been called the underlying ``raw`` object must not be accessed again.
        *Note: Should not normally need to be called explicitly.*
        """
        self._http_response.close()

    def get_size(self) -> int:
        """Return the size of the response in bytes."""
        return int(self._http_response.headers.get("Content-Length", 0))

    def get_content_type(self) -> str:
        """Return the content type of the response."""
        return self._http_response.headers.get("Content-Type", "")

    def get_status_code(self) -> int:
        """Return the status code of the response."""
        return self._http_response.status_code

    def get_headers(self) -> CaseInsensitiveDict[str]:
        """Return the headers of the response."""
        return self._http_response.headers


class TokenAuth(AuthBase):
    """Attach OpenEPD Token Authentication to the given Request object."""

    def __init__(self, token: str) -> None:
        super().__init__()
        self.token = token

    def __eq__(self, other):
        if other is None:
            return False
        if other.__class__ is not self.__class__:
            return False
        return self.token == other.token

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r


class SyncHttpClient:
    """
    HTTP client to communicate with OpenEPD servers via HTTP.

    It works on top of requests library and provides some commonly needed features, such as throttling, retries, etc.
    """

    HTTP_DATE_TIME_FORMAT = "%a, %d %b %Y %H:%M:%S %Z"
    DEFAULT_RETRY_INTERVAL_SEC = 10
    DEFAULT_TIMEOUT_SEC = (15, 2 * 60)

    def __init__(
        self,
        base_url: str,
        throttle_retry_timeout: float | int | datetime.timedelta = 300,
        requests_per_sec: float = 10,
        retry_count: int = 3,
        user_agent: str | None = None,
        timeout_sec: float | tuple[float, float] | None = None,
        auth: AuthBase | None = None,
    ):
        """
        Construct BaseApiClient.

        :param base_url: common part that all request URLs start with
        :param throttle_retry_timeout: how long to wait before retrying throttled request.
                                       Either number of seconds or timedelta
        :param requests_per_sec: requests per second
        :param user_agent: user agent to pass along with a request,
            if `None` then underlying library decides which one to pass
        :param timeout_sec: how long to wait for the server to send data before giving up,
            as a seconds (just a single float), or a (connect timeout, read timeout) tuple.
        :param retry_count: count of retries to perform in case of connection error or timeout.
        """
        self._base_url: str = no_trailing_slash(base_url)
        self._throttler = Throttler(rate_per_sec=requests_per_sec)
        self._throttle_retry_timeout: float = (
            float(throttle_retry_timeout)
            if isinstance(throttle_retry_timeout, (float, int))
            else throttle_retry_timeout.total_seconds()
        )
        self.user_agent = user_agent
        self.timeout = timeout_sec or self.DEFAULT_TIMEOUT_SEC
        self._session: Session | None = None
        self._auth: AuthBase | None = auth
        self._retry_count: int = retry_count

        self._http_retry_handlers: dict[int, RetryHandler] = {}
        self._http_error_handlers: dict[int, ErrorHandler] = {}

        self.register_error_handler(400, DefaultOpenApiErrorHandlers.handle_bad_request)
        self.register_error_handler(401, DefaultOpenApiErrorHandlers.handle_unauthorized)
        self.register_error_handler(403, DefaultOpenApiErrorHandlers.handle_access_denied)
        self.register_error_handler(404, DefaultOpenApiErrorHandlers.handle_not_found)
        self.register_error_handler(500, DefaultOpenApiErrorHandlers.handle_server_error)

    @property
    def base_url(self) -> str:
        """Return base URL for all requests."""
        return self._base_url

    @base_url.setter
    def base_url(self, new_value: str):
        """Set base URL for all requests."""
        self._base_url = no_trailing_slash(new_value)

    @property
    def default_headers(self) -> dict[str, str]:
        """Default headers for requests. Implement if required."""
        headers = {}
        if self.user_agent:
            headers["user-agent"] = self.user_agent
        return headers

    def reset_session(self) -> None:
        """Reset current session (if any). This will clear all cookies and other session data."""
        self._session = None

    def read_bytes_from_url(self, url: str, method: str = "get", **kwargs) -> bytes:
        """
        Perform query to the given endpoint and returns response body as bytes.

        This method will load the ENTIRE CONTENT IN MEMORY. DO NOT USE it to download files. Consider using
        `write_to_stream` instead.
        The request will be performed in the context of API client, default error handling will be applied.

        :param url: url pointing the target endpoint
        :param method: optional HTTP method
        :param kwargs: any other arguments supported by _do_request. See `BaseApiClient._do_request`.
        :return: response content as bytes
        """
        r = self.do_request(method, url, **kwargs)
        content = r.content
        return content

    def read_url_write_to_stream(
        self, url: str, target_stream: IO[bytes], method: str = "get", chunk_size: int = 1024, **kwargs
    ) -> int:
        """
        Perform query to the given endpoint and writes response body to the given stream.

        :param url: url pointing the target endpoint
        :param target_stream: stream to write response body to
        :param method: optional HTTP method
        :param chunk_size: size of the chunk to read from the response
        :param kwargs: any other arguments supported by _do_request. See `BaseApiClient._do_request`.
        :return: number of bytes written
        """
        with self.do_request(method, url, stream=True, **kwargs) as r:
            size = 0
            for chunk in r.iter_content(chunk_size=chunk_size):
                target_stream.write(chunk)
                size += len(chunk)
            return size

    def read_stream_from_url(self, url: str, method: str = "get", **kwargs) -> HttpStreamReader | IO[bytes]:
        """
        Perform query to the given endpoint and returns response body as a stream.

        The request will be performed in the context of API client, default error handling will be applied.
        NOTE: Consider using it as a context manager to handle stream close correctly.

        :param url: url pointing the target endpoint
        :param method: optional HTTP method
        :param kwargs: any other arguments supported by _do_request. See `BaseApiClient._do_request`.
        :return: response content as a stream
        """
        r = self.do_request(method, url, stream=True, **kwargs)
        return HttpStreamReader(r)

    def register_retry_handler(self, http_error: int, retry_handler: RetryHandler | None) -> None:
        """
        Register retry handler for the given HTTP error code.

        This allows to override default error handling for the given HTTP error code from the subclass.
        """
        if retry_handler is not None:
            self._http_retry_handlers[http_error] = retry_handler

    def delete_retry_handler(self, http_error: int) -> None:
        """
        Delete retry handler for the given HTTP error code.

        See _register_error_handler for more details.
        """
        if http_error in self._http_retry_handlers:
            del self._http_retry_handlers[http_error]

    def register_error_handler(self, http_error: int, error_handler: ErrorHandler | None) -> None:
        """
        Register retry handler for the given HTTP error code.

        This allows to override default error handling for the given HTTP error code from the subclass.
        """
        if error_handler is not None:
            self._http_error_handlers[http_error] = error_handler

    def _delete_error_handler(self, http_error: int) -> None:
        """
        Delete retry handler for the given HTTP error code.

        See _register_error_handler for more details.
        """
        if http_error in self._http_error_handlers:
            del self._http_error_handlers[http_error]

    def _run_throttled_request(
        self,
        method: str,
        url: str,
        request_kwargs: dict[str, Any],
        session: Session | None = None,
    ) -> Response:
        left_time = self._throttle_retry_timeout
        # override current session to do request to some other server from the same API client
        session = session or self._current_session
        while True:
            with self._throttler.throttle():
                resp = session.request(method, url, **request_kwargs)
                if resp.status_code == requests_codes.too_many_requests:
                    timeout = self._get_timeout_from_retry_after_header(
                        resp.headers.get("Retry-After"), self.DEFAULT_RETRY_INTERVAL_SEC
                    )
                    if timeout > left_time:
                        return resp
                    logger.info("`%s %s` has been throttled for %s second(s)", method, url, timeout)
                    time.sleep(timeout)
                    left_time -= timeout
                    if left_time > 0:
                        continue
                return resp

    def do_request(
        self,
        method: str,
        endpoint: str,
        params=None,
        data=None,
        json=None,
        files=None,
        headers=None,
        session: Session | None = None,
        auth: AuthBase | None = None,
        raise_for_status: bool = True,
        **kwargs,
    ) -> Response:
        """
        Perform request to the given endpoint.

        See https://requests.readthedocs.io/en/master/api/#requests.request for more details on kwargs.
        """
        headers = headers or self.default_headers

        self._on_before_do_request()

        url = self._get_url_for_request(endpoint)

        request_kwargs = dict(
            params=params,
            data=data,
            json=json,
            files=files,
            headers=headers,
            timeout=self.timeout,
            auth=auth or self._auth,
        )
        request_kwargs.update(kwargs)

        do_request = self._handle_service_unavailable(
            method,
            url,
            self._retry_count,
            partial(self._run_throttled_request, method, url, request_kwargs, session=session),
        )

        response = do_request()

        if response.ok:
            return response

        retry_handler = self._http_retry_handlers.get(response.status_code, None)
        if retry_handler:
            result = retry_handler(do_request)
            response = result or response

        error_handler = self._http_error_handlers.get(response.status_code, None)
        if error_handler:
            response = error_handler(response, raise_for_status)

        if response.ok or not raise_for_status:
            return response

        response.raise_for_status()
        # This can't be handled by static checker because of the dynamic nature of the raise_for_status method
        raise RuntimeError("This line should never be reached")

    def _get_url_for_request(self, path_or_url: str) -> str:
        """
        Generate url for given input.

        If absolute path is given it will be returned as is, otherwise the base url will be prepended.
        :param path_or_url: Either absolute url or base path.
        :return: absolute url
        """
        return self._base_url + path_or_url if not path_or_url.startswith("http") else path_or_url

    @property
    def _current_session(self) -> Session:
        if self._session is None:
            self._session = Session()
        return self._session

    def _on_before_do_request(self):
        """
        Perform any actions before request is sent.

        This is a hook that will be called before `do_request`. Can be overridden to check / refresh access tokens.
        """
        pass

    def _get_timeout_from_retry_after_header(self, retry_after: str | None, default: float = 10.0) -> float:
        if retry_after is None:
            return default
        try:
            return float(retry_after)
        except ValueError:
            # This means the value is not at number of seconds but a date, so we parse it
            try:
                date_in_future = datetime.datetime.strptime(retry_after.strip(), self.HTTP_DATE_TIME_FORMAT)
                return (date_in_future - datetime.datetime.utcnow()).total_seconds()
            except ValueError:
                logger.warning("Invalid Retry-After header: %s", retry_after)
                return default

    @staticmethod
    def _handle_service_unavailable(method: str, url: str, retry_count: int, func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            attempts = retry_count
            response = None
            exception = None
            while attempts > 0:
                exception = None
                try:
                    response = func(*args, **kwargs)
                except (requests.exceptions.ConnectionError, ConnectionError, Timeout) as e:
                    exception = e

                if exception or response.status_code == requests_codes.service_unavailable:
                    secs = random.randint(60, 60 * 5)
                    logger.warning(
                        "%s %s is unavailable. Attempts left: %s. Waiting %s seconds...", method, url, attempts, secs
                    )

                    # wait random number of seconds and request again
                    time.sleep(secs)
                    attempts -= 1
                else:
                    break
            if exception:
                raise exception
            return response

        return wrapper


class DefaultOpenApiErrorHandlers:
    class _Error(NamedTuple):
        summary: str
        error_code: str | None
        details: dict[str, list[str]] | None

    @staticmethod
    def _parse_error_response(response: Response) -> _Error:
        error_text = response.text
        validation_errors: dict[str, Any] | None = None
        error_code: str | None = None
        if response.headers.get("content-type") == "application/json":
            json_error = response.json()
            error_text = json_error.get("detail", error_text)
            validation_errors = json_error.get("validation_errors", None)
            if validation_errors is not None:
                error_code = validation_errors.get("code", None)
                if isinstance(validation_errors.get("detail", 0), str):
                    error_text = validation_errors.get("detail", "")
                    validation_errors = dict(msg=validation_errors.get("detail", []))
                if error_code is not None:
                    error_text = f"[{error_code}] {error_text}"
        return DefaultOpenApiErrorHandlers._Error(error_text, error_code, validation_errors)

    @staticmethod
    def handle_bad_request(response: Response, raise_for_status: bool) -> Response | None:
        if raise_for_status:
            error = DefaultOpenApiErrorHandlers._parse_error_response(response)
            raise errors.ValidationError(
                response.status_code,
                error.summary,
                validation_errors=error.details,
                response=response,
                error_code=error.error_code,
            )
        return response

    @staticmethod
    def handle_not_found(response: Response, raise_for_status: bool) -> Response | None:
        if raise_for_status:
            error = DefaultOpenApiErrorHandlers._parse_error_response(response)
            raise errors.ObjectNotFound(response.status_code, error.summary, response, error_code=error.error_code)
        return response

    @staticmethod
    def handle_unauthorized(response: Response, raise_for_status: bool) -> Response | None:
        if raise_for_status:
            error = DefaultOpenApiErrorHandlers._parse_error_response(response)
            raise errors.NotAuthorizedError(response.status_code, error.summary, response, error_code=error.error_code)
        return response

    @staticmethod
    def handle_access_denied(response: Response, raise_for_status: bool) -> Response | None:
        if raise_for_status:
            error = DefaultOpenApiErrorHandlers._parse_error_response(response)
            raise errors.AccessDeniedError(response.status_code, error.summary, response, error_code=error.error_code)
        return response

    @staticmethod
    def handle_server_error(response: Response, raise_for_status: bool) -> Response | None:
        if raise_for_status:
            error = DefaultOpenApiErrorHandlers._parse_error_response(response)
            raise errors.ServerError(response.status_code, error.summary, response, error_code=error.error_code)
        return response


class BaseApiMethodGroup:
    """Base class for API method groups."""

    def __init__(self, client: SyncHttpClient) -> None:
        """
        Construct a method group.

        :param client: HTTP client to use for requests
        """
        super().__init__()
        self._client = client



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from openepd.compat.pydantic import pyd
from openepd.model.base import BaseOpenEpdSchema
from openepd.model.specs.base import BaseOpenEpdSpec
from openepd.model.validation.numbers import RatioFloat


class ConcreteTypicalApplication(BaseOpenEpdSpec):
    """Typical Application for Concrete."""

    fnd: bool | None = pyd.Field(
        description="Foundation. Typically used in direct contact with soil, e.g. footings, piles, mass concrete, "
        "mat foundations, and similar applications.",
        example=True,
        default=None,
    )
    sog: bool | None = pyd.Field(
        description="Slab on Grade. Typically used in continuously supported horizontal "
        "applications e.g. slab on grade, topping slabs, sidewalks, and roadways.",
        example=True,
        default=None,
    )
    hrz: bool | None = pyd.Field(
        description="Elevated Horizontal. Typically used in elevated horizontal applications, either on metal deck or "
        "where soffit formwork must be removed, e.g. post-tension plates, rebar plates, beams and slabs, "
        "waffle slabs.",
        example=True,
        default=None,
    )
    vrt_wall: bool | None = pyd.Field(description="Vertical Wall.", example=True, default=None)
    vrt_column: bool | None = pyd.Field(description="Vertical Column.", example=True, default=None)
    vrt_other: bool | None = pyd.Field(
        description="Vertical Other. Typically used in vertical applications other than "
        "walls or columns, e.g. sloped surfaces where formwork is required "
        "on multiple faces.",
        example=True,
        default=None,
    )
    sht: bool | None = pyd.Field(
        description="Shotcrete. Pneumatically applied, without formwork on all sides.", example=True, default=None
    )
    cdf: bool | None = pyd.Field(
        description="Flowable Fill (CDF). Typically used to fill voids, backfill retaining "
        "walls, as a sub-base, and similar applications. Also called Controlled "
        "Density Fill (CDF) or Controlled Low Strength Materials (CLSM).",
        example=True,
        default=None,
    )
    sac: bool | None = pyd.Field(
        description="Typically used in concrete sidewalks and barrier curbs.", example=True, default=None
    )
    pav: bool | None = pyd.Field(description="Typically used in pervious concrete", example=True, default=None)
    oil: bool | None = pyd.Field(
        description="Concretes for use in creation, maintenance, and decommissioning of "
        "petroleum extraction wells and similar applications. Includes foamed "
        "cement; often called cement in the drilling industry. Differs from "
        "flowable fill and grout in that it contains no sand or other aggregates.",
        example=True,
        default=None,
    )
    grt: bool | None = pyd.Field(
        description="Cement grouting is a slurry that is placed as a flowable liquid. It is "
        "an effective material for filling and strengthening granular soils, "
        "voids in rocks, foundation underpinnings, and other underground voids. "
        "Also called structural grout, these materials typically impart"
        " significant strength to the system",
        example=True,
        default=None,
    )
    ota: bool | None = pyd.Field(
        description="Typical application not covered by other values.", example=True, default=None
    )


class Cementitious(BaseOpenEpdSchema):
    """List of cementitious materials, and proportion by mass."""

    opc: RatioFloat | None = pyd.Field(
        default=None, description="Ordinary Gray Portland Cement", example=0.5, ge=0, le=1
    )
    wht: RatioFloat | None = pyd.Field(default=None, description="White Portland Cement", example=0.5, ge=0, le=1)
    ggbs: RatioFloat | None = pyd.Field(
        default=None, description="Ground Granulated Blast Furnace Slag", example=0.5, ge=0, le=1
    )
    flyAsh: RatioFloat | None = pyd.Field(
        default=None, description="Fly Ash, including types F, CL, and CH", example=0.5, ge=0, le=1
    )
    siFume: RatioFloat | None = pyd.Field(default=None, description="Silica Fume", example=0.5, ge=0, le=1)
    gg45: RatioFloat | None = pyd.Field(
        default=None, description="Ground Glass, 45um or smaller", example=0.5, ge=0, le=1
    )
    natPoz: RatioFloat | None = pyd.Field(default=None, description="Natural pozzolan", example=0.5, ge=0, le=1)
    mk: RatioFloat | None = pyd.Field(default=None, description="Metakaolin", example=0.5, ge=0, le=1)
    CaCO3: RatioFloat | None = pyd.Field(default=None, description="Limestone", example=0.5, ge=0, le=1)
    other: RatioFloat | None = pyd.Field(default=None, description="Other SCMs", example=0.5, ge=0, le=1)



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
from typing import Literal, TypeAlias, overload

from requests import Response

from openepd.api.base_sync_client import BaseApiMethodGroup
from openepd.api.common import StreamingListResponse, paging_meta_from_v1_api
from openepd.api.dto.common import BaseMeta, OpenEpdApiResponse
from openepd.api.dto.meta import PagingMetaMixin
from openepd.api.utils import encode_path_param, remove_none_id_fields
from openepd.model.generic_estimate import (
    GenericEstimate,
    GenericEstimatePreview,
    GenericEstimateRef,
    GenericEstimateWithDeps,
)


class GenericEstimateApi(BaseApiMethodGroup):
    """API methods for Generic Estimates."""

    @overload
    def get_by_openxpd_uuid(self, uuid: str, with_response: Literal[True]) -> tuple[GenericEstimate, Response]: ...

    @overload
    def get_by_openxpd_uuid(self, uuid: str, with_response: Literal[False] = False) -> GenericEstimate: ...

    def get_by_openxpd_uuid(
        self, uuid: str, with_response: bool = False
    ) -> GenericEstimate | tuple[GenericEstimate, Response]:
        """
        Get Generic Estimate by OpenEPD UUID.

        :param uuid: Open xPD UUID
        :param with_response: whether to return just object or with response
        :return: GE or GE with response depending on param with_response
        :raise ObjectNotFound: if Generic Estimate is not found
        """
        response = self._client.do_request("get", f"/generic_estimates/{uuid}")
        if with_response:
            return GenericEstimate.parse_obj(response.json()), response
        return GenericEstimate.parse_obj(response.json())

    @overload
    def post_with_refs(
        self, ge: GenericEstimateWithDeps, with_response: Literal[True]
    ) -> tuple[GenericEstimate, Response]: ...

    @overload
    def post_with_refs(
        self,
        ge: GenericEstimateWithDeps,
        with_response: Literal[False] = False,
    ) -> GenericEstimate: ...

    def post_with_refs(
        self, ge: GenericEstimateWithDeps, with_response: bool = False
    ) -> GenericEstimate | tuple[GenericEstimate, Response]:
        """
        Post an GenericEstimate with references.

        :param ge: GenericEstimate
        :param with_response: return the response object togather with the GenericEstimate
        :param exclude_defaults: If True, fields with default values are excluded from the payload
        :return: GenericEstimate alone, or GenericEstimate with HTTP Response object depending on parameter
        """
        data = ge.to_serializable(exclude_unset=True, by_alias=True)
        # Remove 'id' fields with None values, as 'id' cannot be None
        data = remove_none_id_fields(data)
        response = self._client.do_request(
            "patch",
            "/generic_estimates/post_with_refs",
            json=data,
        )
        content = response.json()
        if with_response:
            return GenericEstimate.parse_obj(content), response
        return GenericEstimate.parse_obj(content)

    @overload
    def create(self, ge: GenericEstimate, with_response: Literal[True]) -> tuple[GenericEstimateRef, Response]: ...

    @overload
    def create(self, ge: GenericEstimate, with_response: Literal[False] = False) -> GenericEstimateRef: ...

    def create(
        self, ge: GenericEstimate, with_response: bool = False
    ) -> GenericEstimateRef | tuple[GenericEstimateRef, Response]:
        """
        Create a Generic Estimate.

        :param ge: Generic Estimate
        :param with_response: return the response object together with the EPD
        :return: Generic Estimate or Generic Estimate with HTTP Response object depending on parameter
        """
        response = self._client.do_request(
            "post",
            "/generic_estimates",
            json=ge.to_serializable(exclude_unset=True, exclude_defaults=True, by_alias=True),
        )
        content = response.json()
        if with_response:
            return GenericEstimateRef.parse_obj(content), response
        return GenericEstimateRef.parse_obj(content)

    @overload
    def edit(self, ge: GenericEstimate, with_response: Literal[True]) -> tuple[GenericEstimateRef, Response]: ...

    @overload
    def edit(self, ge: GenericEstimate, with_response: Literal[False] = False) -> GenericEstimateRef: ...

    def edit(
        self, ge: GenericEstimate, with_response: bool = False
    ) -> GenericEstimateRef | tuple[GenericEstimateRef, Response]:
        """
        Edit a Generic Estimate.

        :param ge: GenericEstimate
        :param with_response: return the response object together with the GE
        :return: GE or GE with HTTP Response object depending on parameter
        """
        ge_id = ge.id
        if not ge_id:
            raise ValueError("The ID must be set to edit a GenericEstimate.")

        response = self._client.do_request(
            "put",
            f"/generic_estimates/{encode_path_param(ge_id)}",
            json=ge.to_serializable(exclude_unset=True, exclude_defaults=True, by_alias=True),
        )
        content = response.json()
        if with_response:
            return GenericEstimateRef.parse_obj(content), response
        return GenericEstimateRef.parse_obj(content)

    @overload
    def list_raw(
        self, page_num: int, page_size: int, with_response: Literal[False] = False
    ) -> list[GenericEstimatePreview]: ...

    @overload
    def list_raw(
        self, page_num: int, page_size: int, with_response: Literal[True]
    ) -> tuple[list[GenericEstimatePreview], Response]: ...

    def list_raw(
        self, page_num: int = 1, page_size: int = 10, with_response: bool = False
    ) -> list[GenericEstimatePreview] | tuple[list[GenericEstimatePreview], Response]:
        """
        List generic estimates.

        :param page_num: page number
        :param page_size: page size
        :param with_response: whether to return just object or with response

        :return: GE or GE with HTTP Response object depending on parameter
        """
        response = self._client.do_request(
            "get",
            "/generic_estimates",
            params=dict(
                page_number=page_num,
                page_size=page_size,
            ),
        )
        data = [GenericEstimatePreview.parse_obj(o) for o in response.json()]
        if with_response:
            return data, response
        return data

    def list(self, page_size: int | None = None) -> StreamingListResponse[GenericEstimatePreview]:
        """
        List GenericEstimates.

        :param page_size: page size, None for default
        :return: streaming list of GEs
        """

        def _get_page(p_num: int, p_size: int) -> GenericEstimateListResponse:
            data_list, response = self.list_raw(page_num=p_num, page_size=p_size, with_response=True)
            return GenericEstimateListResponse(
                payload=data_list, meta=GenericEstimateSearchMeta(paging=paging_meta_from_v1_api(response))
            )

        return StreamingListResponse[GenericEstimatePreview](_get_page, page_size=page_size)


class GenericEstimateSearchMeta(PagingMetaMixin, BaseMeta):
    """Metadata for EPD Search endpoint."""

    pass


GenericEstimateListResponse: TypeAlias = OpenEpdApiResponse[list[GenericEstimatePreview], GenericEstimateSearchMeta]



#
#  Copyright 2025 by C Change Labs Inc. www.c-change-labs.com
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
__all__ = (
    "WoodDeckingRangeV1",
    "WoodFramingRangeV1",
    "PrefabricatedWoodInsulatedPanelsRangeV1",
    "PrefabricatedWoodTrussRangeV1",
    "CompositeLumberRangeV1",
    "DimensionLumberRangeV1",
    "HeavyTimberRangeV1",
    "MassTimberRangeV1",
    "NonStructuralWoodRangeV1",
    "PrefabricatedWoodRangeV1",
    "SheathingPanelsRangeV1",
    "UnfinishedWoodRangeV1",
    "WoodRangeV1",
)

# NB! This is a generated code. Do not edit it manually. Please see src/openepd/model/specs/README.md


from builtins import float

from openepd.compat.pydantic import pyd
from openepd.model.org import OrgRef
from openepd.model.specs.base import BaseOpenEpdHierarchicalSpec
from openepd.model.specs.enums import (
    AllFabrication,
    AllTimberSpecies,
    CompositeLumberFabrication,
    EngineeredTimberSpecies,
    MassTimberFabrication,
    SawnTimberSpecies,
    SheathingPanelsFabrication,
)
from openepd.model.validation.numbers import RatioFloat
from openepd.model.validation.quantity import AmountRangeLengthMm


class WoodDeckingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Dimensional boards for exterior decking.

    Range version.
    """

    _EXT_VERSION = "1.0"


class WoodFramingRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Dimension lumber for light framing.

    Includes solid and finger-jointed lumber. Standard shapes include 2x4, 2x6, and 2x8.

    Range version.
    """

    _EXT_VERSION = "1.0"


class PrefabricatedWoodInsulatedPanelsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Structural insulated panels (SIPs).

    Consist of wood sheet layer(s) combined with (typically foam) insulation layer(s).

    Range version.
    """

    _EXT_VERSION = "1.0"


class PrefabricatedWoodTrussRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Shop-fabricated wood truss.

    Range version.
    """

    _EXT_VERSION = "1.0"


class CompositeLumberRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Shop-fabricated structural Lumber.

    Includes Laminated Strand Lumber (LSL), Parallel Strand Lumber (PSL), and Laminated Veneer Lumber (LVL).

    Range version.
    """

    _EXT_VERSION = "1.0"

    fabrication: list[CompositeLumberFabrication] | None = pyd.Field(default=None, description="")
    timber_species: list[EngineeredTimberSpecies] | None = pyd.Field(default=None, description="")


class DimensionLumberRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Dimension lumber for framing, decking, and other purposes.

    Range version.
    """

    _EXT_VERSION = "1.0"

    timber_species: list[SawnTimberSpecies] | None = pyd.Field(default=None, description="")
    WoodDecking: WoodDeckingRangeV1 | None = None
    WoodFraming: WoodFramingRangeV1 | None = None


class HeavyTimberRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Large format, unfinished natural timber.

    Range version.
    """

    _EXT_VERSION = "1.0"


class MassTimberRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Engineered heavy timber products.

    Includes glue laminated (glulam), cross-laminated timber (CLT), dowel laminated timber (DLT), and nail laminated
    timber (NLT).

    Range version.
    """

    _EXT_VERSION = "1.0"

    fabrication: list[MassTimberFabrication] | None = pyd.Field(default=None, description="")
    timber_species: list[EngineeredTimberSpecies] | None = pyd.Field(default=None, description="")


class NonStructuralWoodRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Non-structural interior and exterior wood products for trim, cabinets, countertops, etc.

    Range version.
    """

    _EXT_VERSION = "1.0"


class PrefabricatedWoodRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Prefabricated wood structural members made primarily from one or more types of wood.

    Includes products made with metallic connectors, insulation, etc. Excludes products where the wood is merely
    decorative.

    Range version.
    """

    _EXT_VERSION = "1.0"

    PrefabricatedWoodInsulatedPanels: PrefabricatedWoodInsulatedPanelsRangeV1 | None = None
    PrefabricatedWoodTruss: PrefabricatedWoodTrussRangeV1 | None = None


class SheathingPanelsRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Wood sheets used for structural sheathing, including plywood and Oriented Strand Board.

    Range version.
    """

    _EXT_VERSION = "1.0"

    fabrication: list[SheathingPanelsFabrication] | None = pyd.Field(default=None, description="")
    wood_board_thickness: AmountRangeLengthMm | None = pyd.Field(default=None, description="")
    timber_species: list[EngineeredTimberSpecies] | None = pyd.Field(default=None, description="")


class UnfinishedWoodRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Raw logs and other unfinished wood products.

    Range version.
    """

    _EXT_VERSION = "1.0"


class WoodRangeV1(BaseOpenEpdHierarchicalSpec):
    """
    Structural Wood Products used in construction.

    Range version.
    """

    _EXT_VERSION = "1.0"

    forest_practices_certifiers: list[OrgRef] | None = None
    timber_species: list[AllTimberSpecies] | None = pyd.Field(default=None, description="Timber species")
    fabrication: list[AllFabrication] | None = pyd.Field(default=None, description="Timber fabrication")
    weather_exposed: bool | None = pyd.Field(default=None, description="Weather exposed")
    fire_retardant: bool | None = pyd.Field(default=None, description="Fire retardant")
    decay_resistant: bool | None = pyd.Field(default=None, description="Decay resistant")
    fsc_certified: RatioFloat | None = pyd.Field(
        default=None, description="Forest Stewardship Council certified proportion"
    )
    fsc_certified_z: float | None = pyd.Field(default=None, description="")
    recycled_content: RatioFloat | None = pyd.Field(default=None, description="Recycled content")
    recycled_content_z: float | None = pyd.Field(default=None, description="")
    CompositeLumber: CompositeLumberRangeV1 | None = None
    DimensionLumber: DimensionLumberRangeV1 | None = None
    HeavyTimber: HeavyTimberRangeV1 | None = None
    MassTimber: MassTimberRangeV1 | None = None
    NonStructuralWood: NonStructuralWoodRangeV1 | None = None
    PrefabricatedWood: PrefabricatedWoodRangeV1 | None = None
    SheathingPanels: SheathingPanelsRangeV1 | None = None
    UnfinishedWood: UnfinishedWoodRangeV1 | None = None
